/* $Id: keyberry.cpp,v 1.92 2003/01/25 00:17:45 chrisf Exp $ */

/*
 * keyberry - A keyboard driver for the RIM pager
 * Copyright  2000 Chris Faherty <rallymonkey@bellsouth.net>
 *
 * Lots and lots of code copied from:
 * RIM SDK samples
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "Pager.h"
#include "ui32.h"
#include "options.h"
#include <string.h>
#include "peripheral.h"
#include "keyberry.h"

#define VER_MAJOR 1
#define VER_MINOR 3
#define VER_BUILD 10

char VersionPtr[]="Keyberry";
int AppStackSize=4000;

#ifdef BB20
#include "fx100_16x16.h"
#include "fx100_28x28.h"
#endif

TASK mytask;
MESSAGE keymsg;
long idle;
bool commopen;

UIEngine g_uiEngine;

KeyberryScreen *g_kbscreen;

/* these I just made up, not following any standard */
#define KB_LEFT 1
#define KB_RIGHT 2
#define KB_UP 3
#define KB_DOWN 4
#define KB_LSHIFT 5
#define KB_RSHIFT 6
#define KB_FUNC 7
#define KB_BACKSPACE 8
#define KB_TAB 9
#define KB_COMMAND 11
#define KB_SHORTCUT 12
#define KB_ENTER 13
#define KB_CTL 14
#define KB_CAPLOCK 15
#define KB_MENU 16
#define KB_PGUP 17
#define KB_PGDN 18
#define KB_INSERT 19
#define KB_DATE 20
#define KB_ADDRESS 21
#define KB_TODO 22
#define KB_MEMO 23
#define KB_NUMLOCK 24
#define KB_ALT 25
#define KB_ESC 27
#define KB_DELETE 127
#define KB_F1 128
#define KB_F2 129
#define KB_F3 130
#define KB_F4 131
#define KB_F5 132
#define KB_F6 133
#define KB_F7 134
#define KB_F8 135
#define KB_F9 136
#define KB_F10 137
#define KB_WIN 138
#define KB_HOME 139
#define KB_END 140

const unsigned char gotype[]={
    0, '`',
    /* 0x02 */
    '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '-', '=', KEY_BACKSPACE,
    /* 0x0f */
    0x09, /* tab */
    /* 0x10 */
    'q', 'w', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p', '[', ']', '\\',
    /* 0x1d */
    'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', ';', '\'',
    /* 0x28 */
    KEY_ENTER,
    /* 0x29 */
    'z', 'x', 'c', 'v', 'b', 'n', 'm', ',', '.', '/', ' ',
    /* 0x34 */
    0, 0, 0, KB_LEFT, 0, 0, KB_UP, KB_DOWN, 0, 0, KB_RIGHT,
    /* 0x3f */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    /* 0x5b */
    KB_LSHIFT, KB_RSHIFT, KB_ALT, 0, KB_COMMAND,
    /* 0x60 */
    0,0,0,0,0,0,
    /* 0x66 */
    KB_SHORTCUT
};

const unsigned char shift_nums[]={
    ')', '!', '@', '#', '$', '%', '^', '&', '*', '('
};

const unsigned char happyhack[]={
    0,     KB_F9,  0,      KB_F5,  KB_F3,  KB_F1,  KB_F2,   0,
    0,     KB_F10, KB_F8,  KB_F6,  KB_F4,  0x09,   '`',     0,
    0,     0,      KB_LSHIFT, 0,   KB_CTL, 'q',    '1',     0,
    0,     0,      'z',    's',    'a',    'w',    '2',     KB_COMMAND,
    0,     'c',    'x',    'd',    'e',    '4',    '3',     0,
    0,     ' ',    'v',    'f',    't',    'r',    '5',     KB_ALT,
    0,     'n',    'b',    'h',    'g',    'y',    '6',     0,
    0,     0,      'm',    'j',    'u',    '7',    '8',     0,
    0,     ',',   'k',    'i',    'o',    '0',    '9',     0,
    0,     '.',    '/',    'l',    ';',    'p',    '-',     0,
    0,     0,      '\'',    0,      '[',    '=',    0,       0,
    KB_CAPLOCK, KB_RSHIFT, KEY_ENTER, ']', 0, '\\', 0,      0,
    0,     0,      0,      0,      0,      0,      KEY_BACKSPACE, 0,
    0,     KB_END, 0,      KB_LEFT,  KB_HOME, 0,     0,       0,
    0,     KB_DELETE, KB_DOWN, 0,    KB_RIGHT, KB_UP,  KB_DOWN, 0,
    0, 0,     KB_PGDN,    0,     0,       KB_PGUP, 0, 0,
    0,     0,      0,      KB_F7,  0,      0,      0,       0
};

const unsigned char ppk[]={
    '1', '2', '3', 'z', '4', '5', '6', '7',
    KB_COMMAND, 'q', 'w', 'e', 'r', 't', 'y', '`',
    'x', 'a', 's', 'd', 'f', 'g', 'h', ' ',
    KB_CAPLOCK, KB_TAB, KB_CTL, 0, 0, 0, 0, 0,
    0, 0, KB_SHORTCUT, KB_ALT, 0, 0, 0, 0,
    0, 0, 0, 0, 'c', 'v', 'b', 'n',
    '-', '=', KEY_BACKSPACE, KB_F1, '8', '9', '0', ' ',
    '[', ']', '\\', KB_F2, 'u', 'i', 'o', 'p',
    '\'', KEY_ENTER, KB_F3, 0, 'j', 'k', 'l', ';',
    '/', KB_UP, KB_F4, 0, 'm', ',', '.', KEY_ENTER,
    KB_DELETE, KB_LEFT, KB_DOWN, KB_RIGHT, 0, 0, 0, 0,
    KB_LSHIFT, KB_RSHIFT
};

const unsigned char flexis[]={
/* 0x00 */ 'a', 's', 'd', 'f', 'h', 'g', 'z', 'x',
/* 0x08 */ 'c', 'v', KB_WIN, 'b', 'q', 'w', 'e', 'r',
/* 0x10 */ 'y', 't', '1', '2', '3', '4', '6', '5',
/* 0x18 */ '=', '9', '7', '-', '8', '0', ']', 'o',
/* 0x20 */ 'u', '[', 'i', 'p', KEY_ENTER, 'l', 'j', '\'',
/* 0x28 */ 'k', ';', '\\', ',', '/', 'n', 'm', '.',
/* 0x30 */ KB_TAB, ' ', '`', KEY_BACKSPACE, 0, 0, 0, KB_SHORTCUT,
/* 0x38 */ KB_LSHIFT, KB_CAPLOCK, KB_ALT, KB_CTL, 0, KB_COMMAND, 0, KB_RSHIFT,
/* 0x40 */ 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x48 */ 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x50 */ 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x58 */ 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x60 */ 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x68 */ 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x70 */ 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x78 */ 0, 0, KB_DELETE, KB_LEFT, KB_RIGHT, KB_DOWN, KB_UP, 0
};

const unsigned char pocketop[]={
/* 0x00 */ 0, 'q', 0, 'z', 0, 0, 0, 'a',
/* 0x08 */ 0, 'w', 0, 'x', 0, 0, 0, 's',
/* 0x10 */ 0, 'e', 0, 'c', 0, 0, 0, 'd',
/* 0x18 */ 0, 'r', 0, 'v', 0, 0, 0, 'f',
/* 0x20 */ 0, 't', 0, 'b', 0, 0, 0, 'g',
/* 0x28 */ 0, 'y', 0, ' ', 0, 0, 0, 'h',
/* 0x30 */ 0, 'u', 'j', 'n', ' ', 0, 0, 0,
/* 0x38 */ 0, 'i', 'k', 'm', KB_FUNC, 0, 0, 0,
/* 0x40 */ KB_DATE, 'o', 'l', ',', KB_LEFT, 0, 0, 0,
/* 0x48 */ KB_ADDRESS, 'p', ';', '.', KB_DOWN, 0, 0, 0,
/* 0x50 */ KB_TODO, '-', '\'', KB_UP, KB_RIGHT, 0, 0, 0,
/* 0x58 */ KB_MEMO, KB_BACKSPACE, KB_ENTER, KB_DELETE, 0, 0, 0, 0,
/* 0x60 */ 0, 0, 0, 0, 0, KB_LSHIFT, KB_ALT, 0,
/* 0x68 */ 0, KB_ESC, 0, 0, 0, KB_CTL, KB_NUMLOCK, 0,
/* 0x70 */ 0, 0, 0, 0, 0, KB_COMMAND, KB_SHORTCUT, KB_CAPLOCK,
/* 0x78 */ 0, 0, 0, 0, 0, KB_RSHIFT, 0, 0
};

int pt_shifts;
#define PT_SHIFTS_PUNCT 1
#define PT_SHIFTS_NUMLOCK 2
#define PT_SHIFTS_FUNC 4

unsigned char lastch;
long lastticks, kdelay;
bool ppk_repeat;
int ppk_repeat_count;
bool isbreak;

/* Static function prototypes */
static void do_gotype(void);
static void do_happyhack(void);
static void do_ppk(void);
static void do_flexis(void);
static void do_pocketop(void);
static void do_com1(void);

/*-------------------------------------------------------------------------*
 * MenuScreen::HandleMenu
 *
 * This method drives the menu screens attached to both MessageScreen and
 * ComposeScreen. All UI functions to do with the menu are handled here.
 *
 * If the menu is cleared with no operation for the calling screen, -1 is
 * returned. Otherwise, the index of the menu choice is returned. Note that
 * the BACKSPACE hides the menu.
 *-------------------------------------------------------------------------*/
int MenuScreen::HandleMenu(void)
{
    RESULT result;
    MESSAGE msg;
  
    /* Display the menu */
    uiEngine.ProcessMenu(menu);

    for (;;) {
        RimGetMessage(&msg);

        switch (msg.Device) {
        case DEVICE_KEYPAD:
            /* Let the UI process the message */
            result = uiEngine.HandleInput(msg);

            switch (result) {
            case CLICKED:
                /*
                 * A menu item was selected. Erase the menu and return the
                 * selected item.
                 */
                uiEngine.ClearMenu();
                return menu.GetSelectedIndex();

            case UNHANDLED:
                /* Check for BACKSPACE */
                if ((msg.Event == KEY_DOWN) &&
                    (msg.SubMsg == KEY_BACKSPACE)) {
                    uiEngine.ClearMenu();
                    return -1;
                } 
                break;

            case HIDE_MENU:
                /* Erase the menu and return. */
                uiEngine.ClearMenu();
                return -1;
            case UI_SWITCH_TO_BACKGROUND:
                /* Erase the menu and return. */
                uiEngine.ClearMenu();
                return -2;
            default:
                break;
            }
            break;

        }

    }
}

void KeyberryScreen::Display(void)
{
    RESULT result;
    MESSAGE msg;
    int menuIndex;

    /*
     * This will be our message that we send to the foreground app.  We
     * only clear it here because we keep the shift state in it.
     */
    memset(&keymsg, 0, sizeof(MESSAGE));
    keymsg.Device = DEVICE_KEYPAD;

    /* Show a Close menu item only for Blackberry */
#ifndef RIM_HAS_RIBBON
    menu.HideItem(KEYBERRY_MENU_CLOSE);
#endif

    /* Display the screen */
    uiEngine.ProcessScreen(*this);

    for (;;) {
        RimGetMessage(&msg);

        switch (msg.Device) {
        case DEVICE_KEYPAD:

            /* check these keys before the UI does */
            if ((msg.Event == KEY_DOWN) || (msg.Event == KEY_STATUS)) {
                switch (msg.SubMsg) {
                case KEY_ENTER:
                    break;
                #ifdef BB20
                case KEY_ESCAPE:
                    RibbonShowRibbon();
                    break;
                #endif
                }
            }

            /* Let the UI process the message */
            result = uiEngine.HandleInput(msg);

            switch (result) {
            case CLICKED:
                /*
                 * Call the menu. menuIndex contains either -1 or the selected
                 * menu item. Note that item numbers are not reorded when other
                 * items are hidden.
                 */

                if (commopen) {
                    menu.HideItem(KEYBERRY_MENU_ON);
                    menu.ShowItem(KEYBERRY_MENU_OFF);
                    menu.SetSelectedIndex(KEYBERRY_MENU_OFF);
                } else {
                    menu.ShowItem(KEYBERRY_MENU_ON);
                    menu.HideItem(KEYBERRY_MENU_OFF);
                    menu.SetSelectedIndex(KEYBERRY_MENU_ON);
                }

                menuIndex = HandleMenu();

                switch (menuIndex) {
                case KEYBERRY_MENU_ON:
                    /* let's clear shift states etc, whenever opened */
                    memset(&keymsg, 0, sizeof(MESSAGE));
                    keymsg.Device = DEVICE_KEYPAD;
                    switch (m_keytype.GetSelectedIndex()) {
                    case KT_GOTYPE:
                        if (CommOpenPort(2400, 8, COMM_NO_PARITY, 1, 1, 1)) {
                            CommSetDsr(1);
                            commopen = TRUE;
                        }
                        break;
                    case KT_HAPPYHACK:
                        if (CommOpenPort(4800, 8, COMM_NO_PARITY, 1, 1, 1)) {
                            CommSetDsr(1);
                            commopen = TRUE;
                        }
                        break;
                    case KT_PPK:
                        if (CommOpenPort(9600, 8, COMM_NO_PARITY, 1, 1, 1)) {
                            /*
                             * The PPK requires the 3.3V (pin 2) to arrive
                             * first, before the DSR.  But this is hard on
                             * DongleBerry because it uses the DSR signal to
                             * provide the 3.3V (voltage reduced with diodes).
                             * So rather than putting in a switch or a little
                             * battery, we just added a 100uF cap to the 3.3V
                             * signal.  Then we power it up, turn off power,
                             * and finally power it up again while raising
                             * DSR after a time delay.  The concept being that
                             * the cap holds the charge for the 3.3V during
                             * the DSR delayed on.
                             */
                            CommSetDsr(1);
                            RimSleep(50); /* charging the capacitor */
                            CommSetDsr(0);
                            RimSleep(15); /* 3.3V should still be there */
                            CommSetDsr(1); /* finally raise DSR */
                            commopen = TRUE;
                        }
                        break;
                    case KT_FLEXIS:
                        if (CommOpenPort(9600, 8, COMM_NO_PARITY, 1, 1, 1)) {
                            CommSetDsr(1);
                            #ifdef BB21
                            PeripheralPowerSupply(PER_POWER_ON); /* VAUX */
                            #endif
                            commopen = TRUE;
                        }
                        break;
                    case KT_POCKETOP:
                        if (CommOpenPort(9600, 8, COMM_NO_PARITY, 1, 1, 1)) {
                            CommSetDsr(1);
                            commopen = TRUE;
                            pt_shifts = 0;
                        }
                        break;
                    }
                    idle = RimGetTicks(); /* reset the idle time */
                    RimSetTimer((DEVICE_TIMER<<8)+2, 500, TIMER_PERIODIC);
                    break;
                case KEYBERRY_MENU_OFF:
                    if (commopen) {
                        #ifdef BB21
                        PeripheralPowerSupply(PER_POWER_OFF); /* VAUX */
                        #endif
                        CommSetDsr(0);
                        CommClosePort();
                        commopen = FALSE;
                        RimSetTimer((DEVICE_TIMER<<8)+1, 0, TIMER_PERIODIC);
                        RimSetTimer((DEVICE_TIMER<<8)+2, 0, TIMER_PERIODIC);
                    }
                    break;
                #ifdef RIM_HAS_RIBBON
                case KEYBERRY_MENU_CLOSE:
                    RibbonShowRibbon();
                    break;
                #endif
                case -2: /* switching to background while menu is active */
                    /* let's clear the test area */
                    test.Delete(0, test.GetStringLength());
                    break;
                }

                /*
                 * Force the UI to update the display. The menu clears itself,
                 * but does not redraw the bitmap behind...
                 */
                uiEngine.ProcessScreen(*this); 
                break;

            case UNHANDLED:
                break;
            }
            break;

        case DEVICE_SYSTEM:
            switch (msg.Event) {
            case SWITCH_BACKGROUND:
                /* let's clear the test area */
                test.Delete(0, test.GetStringLength());
                break;
            case SWITCH_FOREGROUND:
                break;
            }
            break;

        }
    }
}

static void do_gotype(void)
{
    TASK fore;
    int b;
    unsigned char ch;
    bool isbreak;
    long nowticks;
    bool b1;

    while ((b = CommReadChar()) >= 0) {
        ch = b;
        /* is it a break code */
        if (ch & 0x80) {
            isbreak = TRUE;
            ch &= 0x7f; /* so we can still case it out below */
        } else isbreak = FALSE;
        if (ch >= sizeof(gotype)) continue;
        ch = gotype[ch];
        keymsg.Event = KEY_DOWN;
        switch (ch) {
//        /*
        case KB_LEFT:
            if (!isbreak) {
                keymsg.Event = THUMB_ROLL_UP;
                keymsg.Data[0] |= ALT_STATUS;
            } else keymsg.Data[0] &= ~ALT_STATUS;
            break;
        case KB_RIGHT:
            if (!isbreak) {
                keymsg.Event = THUMB_ROLL_DOWN;
                keymsg.Data[0] |= ALT_STATUS;
            } else keymsg.Data[0] &= ~ALT_STATUS;
            break;
        case KB_UP:
            keymsg.Event = THUMB_ROLL_UP;
            break;
        case KB_DOWN:
            keymsg.Event = THUMB_ROLL_DOWN;
            break;
//        */
        case KB_LSHIFT:
        case KB_RSHIFT:
            keymsg.Event = KEY_STATUS;
            if (isbreak) keymsg.Data[0] &= ~SHIFT_STATUS;
            else keymsg.Data[0] |= SHIFT_STATUS;
            break;
        case KB_ALT:
            keymsg.Event = KEY_STATUS;
            if (isbreak) keymsg.Data[0] &= ~ALT_STATUS;
            else keymsg.Data[0] |= ALT_STATUS;
            break;
        case KB_DELETE:
            ch = KEY_BACKSPACE;
            if (!isbreak) keymsg.Data[0] |= ALT_STATUS;
            else keymsg.Data[0] &= ~ALT_STATUS;
            break;
        case KB_COMMAND:
            keymsg.Event = THUMB_CLICK;
            break;
        case KB_SHORTCUT:
            ch = 0;
            break;
        }

        if (isbreak) {
            /*
             * Is it the break for the last character?  You can
             * chord the keys and you wouldn't want to indicate
             * that all keys are up if they aren't.
             */
            if (ch == lastch) lastch = '\0';
            continue;
        }

        /* repeat character */
        nowticks = RimGetTicks();
        b1 = (nowticks-lastticks) >= kdelay;
        if (ch == 0) continue;
        if ((ch == lastch) && !b1) continue;
        if (ch != lastch) kdelay = 25; else kdelay = 10;
        lastch = ch;
        keymsg.Data[1] = (DWORD) nowticks;

        /* make uppercase characters */
        if (keymsg.Data[0] & SHIFT_STATUS) {
            /* alpha */
            ch = UCASE(ch);
            /* the numbers */
            if ((ch >= '0') && (ch <= '9')) {
                ch = shift_nums[ch-'0'];
            }
            /* other chars */
            switch (ch) {
            case '-': ch = '_'; break;
            case '=': ch = '+'; break;
            case '\\': ch = '|'; break;
            case ';': ch = ':'; break;
            case '\'': ch = '"'; break;
            case ',': ch = '<'; break;
            case '.': ch = '>'; break;
            case '/': ch = '?'; break;
            case '[': ch = '{'; break;
            case ']': ch = '}'; break;
            case '`': ch = '~'; break;
            }
        }

        if (keymsg.Event == KEY_DOWN) keymsg.SubMsg = ch;
        else if ((keymsg.Event == THUMB_ROLL_UP) ||
            (keymsg.Event == THUMB_ROLL_DOWN)) keymsg.SubMsg = 1;
        if ((fore = RimGetForegroundApp())) {
            RimPostMessage(fore, &keymsg);
        }
        /*
         * I get the ticks now because RimPostMessage/RimTaskYield might take
         * a long time and I don't want it to affect the typematic delay.
         */
        RimTaskYield();
        lastticks = RimGetTicks();
    }
}

static void do_happyhack(void)
{
    TASK fore;
    int b;
    unsigned char ch;
    long nowticks;

    while ((b = CommReadChar()) >= 0) {
        ch = b;
        /* weird character, not sure of its purpose */
        if (ch == 0xe0) continue;
        /* is it a break code */
        if (ch == 0xf0) {
            isbreak = TRUE;
            continue; /* the next char is the one that is being lifted */
        } else {
            if (ch >= sizeof(happyhack)) continue;
            ch = happyhack[ch];
        }
        keymsg.Event = KEY_DOWN;
        switch (ch) {
        /*
        case KB_LEFT:
            keymsg.Event = THUMB_ROLL_UP;
            keymsg.Data[0] |= ALT_STATUS;
            break;
        case KB_RIGHT:
            keymsg.Event = THUMB_ROLL_DOWN;
            keymsg.Data[0] |= ALT_STATUS;
            break;
        case KB_UP:
            keymsg.Event = THUMB_ROLL_UP;
            break;
        case KB_DOWN:
            keymsg.Event = THUMB_ROLL_DOWN;
            break;
        */
        case KB_LSHIFT:
        case KB_RSHIFT:
            keymsg.Event = KEY_STATUS;
            if (isbreak) keymsg.Data[0] &= ~SHIFT_STATUS;
            else keymsg.Data[0] |= SHIFT_STATUS;
            break;
        case KB_ALT:
            keymsg.Event = KEY_STATUS;
            if (isbreak) keymsg.Data[0] &= ~ALT_STATUS;
            else keymsg.Data[0] |= ALT_STATUS;
            break;
        case KB_DELETE:
            ch = KEY_BACKSPACE;
            if (!isbreak) keymsg.Data[0] |= ALT_STATUS;
            else keymsg.Data[0] &= ~ALT_STATUS;
            break;
        case KB_COMMAND:
            keymsg.Event = THUMB_CLICK;
            break;
        case KB_SHORTCUT:
            ch = 0;
            break;
        }

        if (isbreak) {
            isbreak = FALSE;
            continue;
        }

        if (ch == 0) continue;
        nowticks = RimGetTicks();
        keymsg.Data[1] = (DWORD) nowticks;

        /* make uppercase characters */
        if (keymsg.Data[0] & SHIFT_STATUS) {
            /* alpha */
            ch = UCASE(ch);
            /* the numbers */
            if ((ch >= '0') && (ch <= '9')) {
                ch = shift_nums[ch-'0'];
            }
            /* other chars */
            switch (ch) {
            case '-': ch = '_'; break;
            case '=': ch = '+'; break;
            case '\\': ch = '|'; break;
            case ';': ch = ':'; break;
            case '\'': ch = '"'; break;
            case ',': ch = '<'; break;
            case '.': ch = '>'; break;
            case '/': ch = '?'; break;
            case '[': ch = '{'; break;
            case ']': ch = '}'; break;
            case '`': ch = '~'; break;
            }
        }

        if (keymsg.Event == KEY_DOWN) keymsg.SubMsg = ch;
        else if ((keymsg.Event == THUMB_ROLL_UP) ||
            (keymsg.Event == THUMB_ROLL_DOWN)) keymsg.SubMsg = 1;
        if ((fore = RimGetForegroundApp())) {
            RimPostMessage(fore, &keymsg);
        }
        RimTaskYield();
    }
}

static void do_ppk(void)
{
    TASK fore;
    int b;
    unsigned char ch;
    bool isbreak=FALSE;
    long nowticks;
    bool b1, already_looped=FALSE;

    /*
     * The PPK doesn't issue repeat characters.  So I had to use a timer which
     * will call this function at the repeat interval while a key is being
     * held down.
     */

    while (((b = CommReadChar()) >= 0) || ppk_repeat) {
        if (ppk_repeat && (b < 0)) {
            ch = lastch;
            if (already_looped) break;
        } else {
            ch = b;
            /* is it a break code */
            if (ch & 0x80) {
                isbreak = TRUE;
                ch &= 0x7f; /* so we can still case it out below */
                if (ppk_repeat) {
                    RimSetTimer((DEVICE_TIMER<<8)+1, 0, TIMER_PERIODIC);
                    ppk_repeat = FALSE;
                }
            } else isbreak = FALSE;
            if (ch >= sizeof(ppk)) continue;
            ch = ppk[ch];
        }
        already_looped = TRUE;
        keymsg.Event = KEY_DOWN;
        switch (ch) {
//        /*
        case KB_LEFT:
            if (!isbreak) {
                keymsg.Event = THUMB_ROLL_UP;
                keymsg.Data[0] |= ALT_STATUS;
            } else keymsg.Data[0] &= ~ALT_STATUS;
            break;
        case KB_RIGHT:
            if (!isbreak) {
                keymsg.Event = THUMB_ROLL_DOWN;
                keymsg.Data[0] |= ALT_STATUS;
            } else keymsg.Data[0] &= ~ALT_STATUS;
            break;
        case KB_UP:
            keymsg.Event = THUMB_ROLL_UP;
            break;
        case KB_DOWN:
            keymsg.Event = THUMB_ROLL_DOWN;
            break;
//        */
        case KB_LSHIFT:
        case KB_RSHIFT:
            keymsg.Event = KEY_STATUS;
            if (isbreak) keymsg.Data[0] &= ~SHIFT_STATUS;
            else keymsg.Data[0] |= SHIFT_STATUS;
            break;
        case KB_ALT:
            keymsg.Event = KEY_STATUS;
            if (isbreak) keymsg.Data[0] &= ~ALT_STATUS;
            else keymsg.Data[0] |= ALT_STATUS;
            break;
        case KB_DELETE:
            ch = KEY_BACKSPACE;
            if (!isbreak) keymsg.Data[0] |= ALT_STATUS;
            else keymsg.Data[0] &= ~ALT_STATUS;
            break;
        case KB_COMMAND:
            keymsg.Event = THUMB_CLICK;
            break;
        case KB_SHORTCUT:
            ch = 0;
            break;
        }

        if (isbreak) {
            /*
             * Is it the break for the last character?  You can
             * chord the keys and you wouldn't want to indicate
             * that all keys are up if they aren't.
             */
            if (ch == lastch) lastch = '\0';
            continue;
        }

        /* repeat character */
        nowticks = RimGetTicks();
        b1 = (nowticks-lastticks) >= kdelay;
        if (ch == 0) continue;
        if ((ch == lastch) && !b1) continue;
        if (ch != lastch) kdelay = 25; else kdelay = 10;
        lastch = ch;
        keymsg.Data[1] = (DWORD) nowticks;

        /* make uppercase characters */
        if (keymsg.Data[0] & SHIFT_STATUS) {
            /* alpha */
            ch = UCASE(ch);
            /* the numbers */
            if ((ch >= '0') && (ch <= '9')) {
                ch = shift_nums[ch-'0'];
            }
            /* other chars */
            switch (ch) {
            case '-': ch = '_'; break;
            case '=': ch = '+'; break;
            case '\\': ch = '|'; break;
            case ';': ch = ':'; break;
            case '\'': ch = '"'; break;
            case ',': ch = '<'; break;
            case '.': ch = '>'; break;
            case '/': ch = '?'; break;
            case '[': ch = '{'; break;
            case ']': ch = '}'; break;
            case '`': ch = '~'; break;
            }
        }

        if (keymsg.Event == KEY_DOWN) keymsg.SubMsg = ch;
        else if ((keymsg.Event == THUMB_ROLL_UP) ||
            (keymsg.Event == THUMB_ROLL_DOWN)) keymsg.SubMsg = 1;
        if ((fore = RimGetForegroundApp())) {
            RimPostMessage(fore, &keymsg);
        }
        /*
         * I get the ticks now because RimPostMessage/RimTaskYield might take
         * a long time and I don't want it to affect the typematic delay.
         */
        RimTaskYield();
        lastticks = RimGetTicks();

        /*
         * On the PPK we mess around with timers to get the auto-repeat.  When
         * we get a key-make we start a 10Hz periodic timer, and we turn it off
         * when we get a key-break.  The other logic is the same as the other
         * keyboard functions; i.e. the start delay and repeat delay logic, even
         * though this could probably have been smarter by having two different
         * timer rates.
         */

        if (!isbreak && !ppk_repeat) {
            RimSetTimer((DEVICE_TIMER<<8)+1, 5, TIMER_PERIODIC);
            ppk_repeat_count = 0;
            ppk_repeat = TRUE;
        }

        /*
         * This part sucks a bit.  Since the PPK's repeat is implied, there is
         * no way to know whether it is simply waiting for a break code or if
         * it just died.  Perhaps there is a feedback signal which I can check
         * for, but since the Palm driver does this same technique I can only
         * assume there is not.  Basically, we just protect against a runaway
         * by limiting the repeat to 32 characters.
         */
        if (++ppk_repeat_count >= 32) {
            RimSetTimer((DEVICE_TIMER<<8)+1, 0, TIMER_PERIODIC);
            ppk_repeat = FALSE;
        }

    }
}

static void do_flexis(void)
{
    TASK fore;
    int b;
    unsigned char ch;
    long nowticks;
    unsigned long oldstat;

    while ((b = CommReadChar()) >= 0) {
        ch = b&0x7f;
        /* is it a break code */
        if (!(b&0x80)) { /* high-bit off is break code */
            isbreak = TRUE;
        }
        if (ch >= sizeof(flexis)) continue;
        ch = flexis[ch];
        keymsg.Event = KEY_DOWN;
        switch (ch) {
        case KB_LEFT:
            keymsg.Event = THUMB_ROLL_UP;
            if (!isbreak) {
                keymsg.Data[0] |= ALT_STATUS;
            } else keymsg.Data[0] &= ~ALT_STATUS;
            break;
        case KB_RIGHT:
            keymsg.Event = THUMB_ROLL_DOWN;
            if (!isbreak) {
                keymsg.Data[0] |= ALT_STATUS;
            } else keymsg.Data[0] &= ~ALT_STATUS;
            break;
        case KB_UP:
            keymsg.Event = THUMB_ROLL_UP;
            break;
        case KB_DOWN:
            keymsg.Event = THUMB_ROLL_DOWN;
            break;
        case KB_LSHIFT:
        case KB_RSHIFT:
            keymsg.Event = KEY_STATUS;
            if (isbreak) {
                keymsg.Data[0] &= ~SHIFT_STATUS;
            } else keymsg.Data[0] |= SHIFT_STATUS;
            break;
        case KB_CAPLOCK:
            keymsg.Event = KEY_STATUS;
            /* toggle */
            if (!isbreak) keymsg.Data[0] ^= CAPS_LOCK;
            break;
        case KB_ALT:
            keymsg.Event = KEY_STATUS;
            if (isbreak) keymsg.Data[0] &= ~ALT_STATUS;
            else keymsg.Data[0] |= ALT_STATUS;
            break;
        case KB_DELETE:
            ch = KEY_DELETE;
            break;
        case KB_COMMAND:
            keymsg.Event = THUMB_CLICK;
            break;
        case KB_SHORTCUT:
            ch = 0xff;
            break;
        case KB_WIN:
            ch = KEY_BACKSPACE;
            #ifdef BB20
            ch = KEY_ESCAPE; /* for handheld models */
            #endif
            break;
        }

        if (isbreak) {
            isbreak = FALSE;
            continue;
        }

        if (ch == 0xff) continue;
        nowticks = RimGetTicks();
        keymsg.Data[1] = (DWORD) nowticks;

        #ifdef BB20
        /* check for special launch keys */
        bool spec=FALSE;
        if ((keymsg.Event == KEY_DOWN) && (keymsg.Data[0] & ALT_STATUS)) {
            switch (ch) {
            case 'a': /* Main Menu */
                RibbonShowRibbon();
                spec = TRUE;
                break;
            case '1': /* Address Book */
                RibbonSetNextApplication("Address Book");
                RibbonShowRibbon();
                spec = TRUE;
                break;
            case '2': /* Compose */
                RibbonSetNextApplication("Compose");
                RibbonShowRibbon();
                spec = TRUE;
                break;
            case '3': /* Alarm */
                RibbonSetNextApplication("Alarm");
                RibbonShowRibbon();
                spec = TRUE;
                break;
            case '4': /* Options */
                RibbonSetNextApplication("Options");
                RibbonShowRibbon();
                spec = TRUE;
                break;
            case 'f': /* Search Messages */
                RibbonSetNextApplication("Search Messages");
                RibbonShowRibbon();
                spec = TRUE;
                break;
            case '=': /* Power Off */
                RimRequestPowerOff();
                spec = TRUE;
                break;
            }
        }

        if (spec) continue;
        #endif

        /* make uppercase characters */
        if ((keymsg.Data[0] & CAPS_LOCK) || (keymsg.Data[0] & SHIFT_STATUS)) {
            /* alpha */
            ch = UCASE(ch);
            /* these chars only if shift is active, not capslock */
            if (keymsg.Data[0] & SHIFT_STATUS) {
                /* the numbers */
                if ((ch >= '0') && (ch <= '9')) {
                    ch = shift_nums[ch-'0'];
                }
                /* other chars */
                switch (ch) {
                case '-': ch = '_'; break;
                case '=': ch = '+'; break;
                case '\\': ch = '|'; break;
                case ';': ch = ':'; break;
                case '\'': ch = '"'; break;
                case ',': ch = '<'; break;
                case '.': ch = '>'; break;
                case '/': ch = '?'; break;
                case '[': ch = '{'; break;
                case ']': ch = '}'; break;
                case '`': ch = '~'; break;
                }
            }
        }

        /* temporarily turn capslock into shift state */
        oldstat = keymsg.Data[0];
        if (keymsg.Data[0] & CAPS_LOCK) keymsg.Data[0] |= SHIFT_STATUS;

        if (keymsg.Event == KEY_DOWN) keymsg.SubMsg = ch;
        else if ((keymsg.Event == THUMB_ROLL_UP) ||
            (keymsg.Event == THUMB_ROLL_DOWN)) keymsg.SubMsg = 1;
        if ((fore = RimGetForegroundApp())) {
            RimPostMessage(fore, &keymsg);
        }
        RimTaskYield();

        /* temporarily reset shift state from capslock */
        if (keymsg.Data[0] & CAPS_LOCK) {
            keymsg.Data[0] = oldstat;
        }

    }
}

static void do_pocketop(void)
{
    TASK fore;
    int b;
    unsigned char ch;
    long nowticks;
    unsigned long oldstat;
    bool b1, already_looped=FALSE;

    while (((b = CommReadChar()) >= 0) || ppk_repeat) {
        if (ppk_repeat && (b < 0)) {
            ch = lastch;
            if (already_looped) break;
        } else {
            ch = b&0x7f;
            /* is it a break code */
            if (b&0x80) { /* high-bit on is break code */
                isbreak = TRUE;
                if (ppk_repeat) {
                    RimSetTimer((DEVICE_TIMER<<8)+1, 0, TIMER_PERIODIC);
                    ppk_repeat = FALSE;
                }
            } else isbreak = FALSE;
            if (ch >= sizeof(pocketop)) continue;
            ch = pocketop[ch];
        }
        already_looped = TRUE;
        keymsg.Event = KEY_DOWN;
        switch (ch) {
        case KB_LEFT:
            if (pt_shifts & PT_SHIFTS_NUMLOCK) break;
            keymsg.Event = THUMB_ROLL_UP;
            if (!isbreak) {
                keymsg.Data[0] |= ALT_STATUS;
            } else keymsg.Data[0] &= ~ALT_STATUS;
            break;
        case KB_RIGHT:
            if (pt_shifts & PT_SHIFTS_NUMLOCK) break;
            keymsg.Event = THUMB_ROLL_DOWN;
            if (!isbreak) {
                keymsg.Data[0] |= ALT_STATUS;
            } else keymsg.Data[0] &= ~ALT_STATUS;
            break;
        case KB_UP:
            if (pt_shifts & PT_SHIFTS_NUMLOCK) break;
            keymsg.Event = THUMB_ROLL_UP;
            break;
        case KB_DOWN:
            if (pt_shifts & PT_SHIFTS_NUMLOCK) break;
            keymsg.Event = THUMB_ROLL_DOWN;
            break;
        case KB_LSHIFT:
        case KB_RSHIFT:
            keymsg.Event = KEY_STATUS;
            if (isbreak) {
                keymsg.Data[0] &= ~SHIFT_STATUS;
            } else keymsg.Data[0] |= SHIFT_STATUS;
            break;
        case KB_CAPLOCK:
            keymsg.Event = KEY_STATUS;
            /* toggle */
            if (!isbreak) keymsg.Data[0] ^= CAPS_LOCK;
            break;
        case KB_ALT:
            keymsg.Event = KEY_STATUS;
            if (isbreak) keymsg.Data[0] &= ~ALT_STATUS;
            else keymsg.Data[0] |= ALT_STATUS;
            break;
        case KB_DELETE:
            ch = KEY_BACKSPACE;
            if (!isbreak) keymsg.Data[0] |= ALT_STATUS;
            else keymsg.Data[0] &= ~ALT_STATUS;
            break;
        case KB_COMMAND:
            keymsg.Event = THUMB_CLICK;
            break;
        case KB_SHORTCUT: /* really Punct on pocketop */
            ch = 0xff;
            if (!isbreak) pt_shifts |= PT_SHIFTS_PUNCT;
            else pt_shifts &= ~PT_SHIFTS_PUNCT;
            break;
        case KB_NUMLOCK:
            ch = 0xff;
            if (!isbreak) pt_shifts |= PT_SHIFTS_NUMLOCK;
            else pt_shifts &= ~PT_SHIFTS_NUMLOCK;
            break;
        case KB_FUNC:
            if (pt_shifts & PT_SHIFTS_NUMLOCK) break;
            ch = 0xff;
            if (!isbreak) pt_shifts |= PT_SHIFTS_FUNC;
            else pt_shifts &= ~PT_SHIFTS_FUNC;
            break;
        case KB_WIN:
            ch = KEY_BACKSPACE;
            #ifdef BB20
            ch = KEY_ESCAPE; /* for handheld models */
            #endif
            break;
        }

        if (isbreak) {
            /*
             * Is it the break for the last character?  You can
             * chord the keys and you wouldn't want to indicate
             * that all keys are up if they aren't.
             */
            if (ch == lastch) lastch = '\0';
            continue;
        }

        if (ch == 0xff) continue;

        /* repeat character */
        nowticks = RimGetTicks();
        b1 = (nowticks-lastticks) >= kdelay;
        if (ch == 0) continue;
        if ((ch == lastch) && !b1) continue;
        if (ch != lastch) kdelay = 25; else kdelay = 10;
        lastch = ch;
        keymsg.Data[1] = (DWORD) nowticks;

        /* make uppercase characters */
        if ((keymsg.Data[0] & CAPS_LOCK) || (keymsg.Data[0] & SHIFT_STATUS)) {
            /* alpha */
            ch = UCASE(ch);
            /* these chars only if shift is active, not capslock */
            if (keymsg.Data[0] & SHIFT_STATUS) {
                /* other chars */
                switch (ch) {
                case '-': ch = '_'; break;
                case ';': ch = ':'; break;
                case ',': ch = '<'; break;
                case '.': ch = '>'; break;
                }
            }
        }

        /* punct keys */
        if (pt_shifts & PT_SHIFTS_PUNCT) {
            switch (UCASE(ch)) {
            case 'Q': ch = '!'; break;
            case 'W': ch = '@'; break;
            case 'E': ch = '#'; break;
            case 'R': ch = '$'; break;
            case 'T': ch = '%'; break;
            case 'Y': ch = '^'; break;
            case 'U': ch = '&'; break;
            case 'I': ch = '*'; break;
            case 'O': ch = '('; break;
            case 'P': ch = ')'; break;
            case '-': ch = '+'; break;
            case 'A': ch = '`'; break;
            case 'S': ch = '~'; break;
            case 'D': ch = '/'; break;
            case 'F': ch = '?'; break;
            case 'G': ch = '['; break;
            case 'H': ch = ']'; break;
            case 'J': ch = '{'; break;
            case 'K': ch = '}'; break;
            case 'L': ch = '\\'; break;
            case ';': ch = '|'; break;
            case '\'': ch = '='; break;
            }
        }

        /* numlock keys */
        if (pt_shifts & PT_SHIFTS_NUMLOCK) {
            switch (UCASE(ch)) {
            case 'Q': ch = '1'; break;
            case 'W': ch = '2'; break;
            case 'E': ch = '3'; break;
            case 'R': ch = '4'; break;
            case 'T': ch = '5'; break;
            case 'Y': ch = '6'; break;
            case 'U': ch = '7'; break;
            case 'I': ch = '8'; break;
            case 'O': ch = '9'; break;
            case 'P': ch = '0'; break;
            case '-': ch = '/'; break;
            case 'J': ch = '4'; break;
            case 'K': ch = '5'; break;
            case 'L': ch = '6'; break;
            case ';': ch = '*'; break;
            case 'M': ch = '1'; break;
            case ',': ch = '2'; break;
            case '.': ch = '3'; break;
            case KB_UP: ch = '-'; break;
            case KB_FUNC: ch = '0'; break;
            case KB_LEFT: ch = '.'; break;
            case KB_DOWN: ch = '+'; break;
            case KB_RIGHT: ch = '='; break;
            }
        }

        /* temporarily turn capslock into shift state */
        oldstat = keymsg.Data[0];
        if (keymsg.Data[0] & CAPS_LOCK) keymsg.Data[0] |= SHIFT_STATUS;

        if (keymsg.Event == KEY_DOWN) keymsg.SubMsg = ch;
        else if ((keymsg.Event == THUMB_ROLL_UP) ||
            (keymsg.Event == THUMB_ROLL_DOWN)) keymsg.SubMsg = 1;
        if ((fore = RimGetForegroundApp())) {
            RimPostMessage(fore, &keymsg);
        }

        RimTaskYield();
        lastticks = RimGetTicks();
        
        if (!isbreak && !ppk_repeat) {
            RimSetTimer((DEVICE_TIMER<<8)+1, 5, TIMER_PERIODIC);
            ppk_repeat_count = 0;
            ppk_repeat = TRUE;
        }

        if (++ppk_repeat_count >= 32) {
            RimSetTimer((DEVICE_TIMER<<8)+1, 0, TIMER_PERIODIC);
            ppk_repeat = FALSE;
        }

        /* temporarily reset shift state from capslock */
        if (keymsg.Data[0] & CAPS_LOCK) {
            keymsg.Data[0] = oldstat;
        }

    }
}

static void do_com1(void)
{
    if (!g_kbscreen) return;
    switch (g_kbscreen->m_keytype.GetSelectedIndex()) {
    case KT_GOTYPE:
        do_gotype();
        break;
    case KT_HAPPYHACK:
        do_happyhack();
        break;
    case KT_PPK:
        do_ppk();
        break;
    case KT_FLEXIS:
        do_flexis();
        break;
    case KT_POCKETOP:
        do_pocketop();
        break;
    }
    idle = RimGetTicks(); /* reset the idle time */
}

TASK NewRimGetMessage(MESSAGE *msg)
{
    TASK tsk;
    tsk = RimGetMessage(msg);
    /*
     * Here I fix it up in my applications to allow cursor control from
     * Keyberry.
     *
     * OS < 2.1 does not allow THUMB_ROLL events to be posted properly because
     * it will always inspect the hardware roller to determine the SubMsg
     * field.. thereby making it zero if you post the message artificially.
     *
     * Even worse is that after it reads the hardware, it then decides to
     * turn all zero SubMsg events into THUMB_ROLL_DOWN regardless of whether
     * it was UP to begin with.
     *
     * So to implement cursor movement, I chose to use my own characters.  And
     * then I translate them here into the proper THUMB_ROLL events.
     *
     */
    switch (msg->Device) {
    case DEVICE_KEYPAD:
        switch (msg->Event) {
        case KEY_DOWN:
            switch (msg->SubMsg) {
            case KB_LEFT:
                msg->Data[0] |= ALT_STATUS;
            case KB_UP:
                msg->Event = THUMB_ROLL_UP;
                msg->SubMsg = 1;
                break;
            case KB_RIGHT:
                msg->Data[0] |= ALT_STATUS;
            case KB_DOWN:
                msg->Event = THUMB_ROLL_DOWN;
                msg->SubMsg = 1;
                break;
            }
            break;
        }
        break;
    }

    return tsk;
}

int RollerCallback(MESSAGE *msg)
{
    switch (msg->Device) {
    case DEVICE_KEYPAD:
        switch (msg->Event) {
        case KEY_DOWN:
            switch (msg->SubMsg) {
            case KB_LEFT:
                msg->Data[0] |= ALT_STATUS;
            case KB_UP:
                msg->Event = THUMB_ROLL_UP;
                msg->SubMsg = 1;
                break;
            case KB_RIGHT:
                msg->Data[0] |= ALT_STATUS;
            case KB_DOWN:
                msg->Event = THUMB_ROLL_DOWN;
                msg->SubMsg = 1;
                break;
            }
            break;
        }
        break;
    }

    return 1;
}

/*
 * I stuck all my global housekeeping in one callback function which is
 * registered for multiple event types.
 */
int MultiCallback(MESSAGE *msg)
{
    switch (msg->Device) {
        case DEVICE_SYSTEM:
            switch (msg->Event) {
            case POWER_OFF:
                if (commopen) {
                    #ifdef BB21
                    PeripheralPowerSupply(PER_POWER_OFF); /* VAUX */
                    #endif
                    CommSetDsr(0);
                    CommClosePort();
                    commopen = FALSE;
                    RimSetTimer((DEVICE_TIMER<<8)+1, 0, TIMER_PERIODIC);
                    RimSetTimer((DEVICE_TIMER<<8)+2, 0, TIMER_PERIODIC);
                }
                break;
            }
            break;

        case DEVICE_COM1:
            if (!commopen) break;
            switch (msg->Event) {
            case COMM_RX_AVAILABLE:
                do_com1();
                break;
            }
            break;

        case DEVICE_TIMER:
            if (!commopen || !g_kbscreen) break;
            switch (msg->Event) {
            case (DEVICE_TIMER<<8)+1:
                switch (g_kbscreen->m_keytype.GetSelectedIndex()) {
                case KT_PPK:
                    do_ppk();
                    break;
                case KT_POCKETOP:
                    do_pocketop();
                    break;
                }
                break;
            case (DEVICE_TIMER<<8)+2:
                RimPagerSetup(0x8426ce98, 0); /* mysterious, reset idle time */
                break;
            }
            break;

        case DEVICE_RTC:
            /* this is called automatically every minute */
            if (msg->Event == RTC_CLOCK_UPDATE) {
                if (!commopen) break;
                long t=RimGetTicks();
                /* if idle for more than 15 minutes close the port */
                if ((t-idle) >= 15*60*100) {
                    if (commopen) {
                        #ifdef BB21
                        PeripheralPowerSupply(PER_POWER_OFF); /* VAUX */
                        #endif
                        CommSetDsr(0);
                        CommClosePort();
                        commopen = FALSE;
                        RimSetTimer((DEVICE_TIMER<<8)+1, 0, TIMER_PERIODIC);
                        RimSetTimer((DEVICE_TIMER<<8)+2, 0, TIMER_PERIODIC);
                    }
                }
            }
            break;

        }
    return 1;
}

/*-------------------------------------------------------------------------*
 * PagerMain
 *-------------------------------------------------------------------------*/
void PagerMain(void)
{
    /*
     * Initialize some global variables.  OS 2.1 has a bug whereby globals
     * with initializers will be set as read-only.  So I can't use any
     * initializers on globals or static locals which I need as read/write.
     */
    idle = 0;
    commopen = FALSE;
    g_kbscreen = NULL;
    lastch = '\0';
    lastticks = 0;
    kdelay = 0;
    ppk_repeat = FALSE;
    ppk_repeat_count = 0;
    isbreak = FALSE;
    pt_shifts = 0;

    mytask = RimGetCurrentTaskID();

    /* This shows the version number on the option screen */
    RegisterApplication(VersionPtr, VER_MAJOR, VER_MINOR, VER_BUILD);

    /* Let other applications initialize */
    RimTaskYield();

    /* Start the user interface engine */
    g_uiEngine.Initialize();

    KeyberryScreen keyberryScreen(g_uiEngine);

    /*
     * Make a global pointer to the keyberry screen.  I hate to do this but
     * I need to reference the keyboard choice field, and the alternative
     * is to maintain a synchronized global variable for the keyboard choice.
     */
    g_kbscreen = &keyberryScreen;

#ifdef BB20
    DEVICE_INFO info;
    RimGetDeviceInfo(&info); 
    if (info.deviceType == DEV_PAGER) { 
        RibbonRegisterApplication(VersionPtr, &fx100_16x16_BitMap, 0, 21);
    } else if (info.deviceType == DEV_HANDHELD) {
        RibbonRegisterApplication(VersionPtr, &fx100_28x28_BitMap, 0, 21);
    }
#else
#ifdef RIM_HAS_RIBBON
    RibbonRegisterApplication(VersionPtr, &fx100_16x16_BitMap, 0, 21);
#endif
#endif

    /* I use callbacks for all my global events */
    RimRegisterMessageCallback(POWER_OFF, 0xffffffff, MultiCallback);
    RimRegisterMessageCallback(COMM_RX_AVAILABLE, 0xffffffff, MultiCallback);
    RimRegisterMessageCallback((DEVICE_TIMER<<8)+1, 0xffffffff, MultiCallback);
    RimRegisterMessageCallback((DEVICE_TIMER<<8)+2, 0xffffffff, MultiCallback);
    RimRegisterMessageCallback(DEVICE_RTC<<8, 0xff00, MultiCallback);

    /* Call the main message loop -- never exits. */
    keyberryScreen.Display();

#ifdef RIM_HAS_RIBBON
    RibbonUnregisterApplication(VersionPtr);
#endif

}
