/* $Id: keyberry.h,v 1.23 2003/01/25 00:09:29 chrisf Exp $ */

/*
 * keyberry - A keyboard driver for the RIM pager
 * Copyright  2000 Chris Faherty <rallymonkey@bellsouth.net>
 *
 * Lots and lots of code copied from:
 * RIM SDK samples
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef __KEYBERRY_H
#define __KEYBERRY_H

#ifndef UCASE
#define UCASE(_a) ((_a)>='a' && (_a)<='z' ? (_a)-32:(_a))
#endif

#ifndef RIM950
    #define RIM_HAS_RIBBON
#endif

#ifdef RIM_HAS_RIBBON
    #include "Ribbon.h"
#endif

char const * const MENU_SEPARATOR_BAR = "__";

const int numberKeyberryMenu = 5;
char const * const keyberryMenu[numberKeyberryMenu]={
    MENU_SEPARATOR_BAR,   /* 0 */
    "Keyboard on",        /* 1 */
    "Keyboard off",       /* 2 */
    "Close",              /* 3 */
    MENU_SEPARATOR_BAR    /* 4 */
};

enum {
    KEYBERRY_MENU_SEP_1 = 0,
    KEYBERRY_MENU_ON = 1,
    KEYBERRY_MENU_OFF = 2,
    KEYBERRY_MENU_CLOSE = 3,
    KEYBERRY_MENU_SEP_2 = 0
};

char const * const keyTypes[]={
    "Flexis",
    0, /* eliminate other keyboard types for show */
    "Pocketop",
    "GoType",
    "Happy Hacking",
    "Palm Portable",
    0
};

#define KT_FLEXIS 0
#define KT_POCKETOP 1
#define KT_GOTYPE 2
#define KT_HAPPYHACK 3
#define KT_PPK 4

/*
 * MenuScreen defines a simple screen that contains a menu and a simple message
 * loop to handle the menu.
 */
class MenuScreen : public Screen {
protected:
    UIEngine &uiEngine;
    Menu menu;

public:
    MenuScreen(UIEngine &uiEngine,
        char const * const * const menuItems,
        int const numberItems) :
        menu(menuItems, numberItems),
        uiEngine(uiEngine) {}
    int HandleMenu(void);
};

class KeyberryScreen : public MenuScreen {
public:
    Title title;
    Choice m_keytype;
    Edit test;

    KeyberryScreen(UIEngine &uiEngine) :
        MenuScreen(uiEngine, keyberryMenu, numberKeyberryMenu),
        m_keytype("Keyboard type:", keyTypes),
        test("Test:", 64, 1024)
    {
        title.SetText("Keyberry");
        AddLabel(title);
        AddField(m_keytype);
        AddField(test);
    }
    void Display(void);
};

/* function prototypes */
TASK NewRimGetMessage(MESSAGE *msg);
int TimerCallback(MESSAGE *msg);
int RollerCallback(MESSAGE *msg);
void PagerMain(void);

#endif
